/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_efrag.c

#include "quakedef.h"


/*
===============================================================================

ENTITY FRAGMENT FUNCTIONS

===============================================================================
*/

efrag_t		**ef_lastlink;
vec3_t		r_efmins, r_efmaxs;
entity_t	*r_efaddent;


/*
===================
R_SplitEntityOnNode
===================
*/
void R_SplitEntityOnNode (mnode_t *node)
{
	efrag_t		*ef;
	mplane_t	*splitplane;
	mleaf_t		*leaf;
	int			sides;

	if (node->contents == CONTENTS_SOLID)
		return;

	// add an efrag if the node is a leaf
	if (node->contents < 0)
	{
		leaf = (mleaf_t *) node;

		// grab an efrag off the free list
		ef = Hunk_Alloc (sizeof (efrag_t));

		ef->entity = r_efaddent;

		// add the entity link	
		*ef_lastlink = ef;
		ef_lastlink = &ef->entnext;
		ef->entnext = NULL;

		// set the leaf links
		ef->leaf = leaf;
		ef->leafnext = leaf->efrags;
		leaf->efrags = ef;

		return;
	}

	// NODE_MIXED
	splitplane = node->plane;
	sides = BOX_ON_PLANE_SIDE (r_efmins, r_efmaxs, splitplane);

	// recurse down the contacted sides
	if (sides & 1) R_SplitEntityOnNode (node->children[0]);
	if (sides & 2) R_SplitEntityOnNode (node->children[1]);
}



/*
===========
R_AddEfrags
===========
*/
void R_AddEfrags (entity_t *ent)
{
	int	i;

	if (!ent->model)
		return;

	r_efaddent = ent;
	ef_lastlink = &ent->efrag;

	for (i = 0; i < 3; i++)
	{
		r_efmins[i] = ent->origin[i] + ent->model->mins[i];
		r_efmaxs[i] = ent->origin[i] + ent->model->maxs[i];
	}

	R_SplitEntityOnNode (cl.worldmodel->nodes);
}


/*
================
R_StoreEfrags

// FIXME: a lot of this goes away with edge-based
================
*/
void R_StoreEfrags (efrag_t **ppefrag)
{
	entity_t	*pent;
	model_t		*clmodel;
	efrag_t		*pefrag;

	while ((pefrag = *ppefrag) != NULL)
	{
		pent = pefrag->entity;
		clmodel = pent->model;

		switch (clmodel->type)
		{
		case mod_alias:
		case mod_brush:
		case mod_sprite:
			pent = pefrag->entity;

			if (pent->visframe != r_framecount)
			{
				R_AddVisedict (pent);

				// mark that we've recorded this entity for this frame
				pent->visframe = r_framecount;
			}

			ppefrag = &pefrag->leafnext;
			break;

		default:
			Sys_Error ("R_StoreEfrags: Bad entity type %d\n", clmodel->type);
		}
	}
}


/*
===============================================================================

LIGHT FRAGMENT FUNCTIONS

===============================================================================
*/

lfrag_t		**lf_lastlink;


void R_SplitLightOnNode (staticlight_t *light, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;

	if (node->contents == CONTENTS_SOLID)
		return;

	// add an efrag if the node is a leaf
	if (node->contents < 0)
	{
		mleaf_t *leaf = (mleaf_t *) node;

		// grab an efrag off the free list
		lfrag_t *lf = Hunk_Alloc (sizeof (lfrag_t));

		lf->light = light;

		// add the entity link	
		*lf_lastlink = lf;
		lf_lastlink = &lf->lightnext;
		lf->lightnext = NULL;

		// set the leaf links
		lf->leaf = leaf;
		lf->leafnext = leaf->lfrags;
		leaf->lfrags = lf;

		return;
	}

	// go down contacted sides
	splitplane = node->plane;
	dist = DotProduct (light->origin, splitplane->normal) - splitplane->dist;

	if (dist > light->light)
		R_SplitLightOnNode (light, node->children[0]);
	else if (dist < -light->light)
		R_SplitLightOnNode (light, node->children[1]);
	else
	{
		R_SplitLightOnNode (light, node->children[0]);
		R_SplitLightOnNode (light, node->children[1]);
	}
}


void R_AddLfrags (staticlight_t *light)
{
	lf_lastlink = &light->lfrag;
	R_SplitLightOnNode (light, cl.worldmodel->nodes);
}


void R_StoreLfrags (lfrag_t **pplfrag)
{
	staticlight_t *plight;
	lfrag_t		*plfrag;

	while ((plfrag = *pplfrag) != NULL)
	{
		plight = plfrag->light;

		// mark that we've recorded this light for this frame
		plight->visframe = r_framecount;

		pplfrag = &plfrag->leafnext;
	}
}


