/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking
int			r_dlightframecount;

mplane_t	frustum[4];

int			r_spottexture;		// spotlight falloff
int			playertextures;		// up to 16 color translated skins

// view origin
float	vup[4];
float	vpn[4];
float	vright[4];
float	r_origin[4];

glmatrix	r_world_matrix;
glmatrix	r_projection_matrix;
glmatrix	r_mvp_matrix;


// screen size info
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

texture_t	*r_notexture_mip;

float		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);
void R_AnimateLight (double time);

cvar_t	r_norefresh = {"r_norefresh", "0"};
cvar_t	r_drawentities = {"r_drawentities", "1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel", "1"};
cvar_t	r_fullbright = {"r_fullbright", "0"};
cvar_t	r_lightmap = {"r_lightmap", "0"};
cvar_t	r_shadows = {"r_shadows", "0"};
cvar_t	r_wateralpha = {"r_wateralpha", "1"};
cvar_t	r_dynamic = {"r_dynamic", "1"};
cvar_t	r_novis = {"r_novis", "0"};

cvar_t	gl_finish = {"gl_finish", "0"};
cvar_t	gl_cull = {"gl_cull", "1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels", "1"};
cvar_t	gl_affinemodels = {"gl_affinemodels", "0"};
cvar_t	gl_polyblend = {"gl_polyblend", "1"};
cvar_t	gl_nocolors = {"gl_nocolors", "0"};


void R_SetupAliasModel (entity_t *e);
void R_SetupBrushModel (entity_t *e);

void R_AddVisedict (entity_t *e)
{
	if (cl_numvisedicts < MAX_VISEDICTS)
	{
		// calc the proper transform for the entity
		switch (e->model->type)
		{
		case mod_alias:
			R_SetupAliasModel (e);
			break;

		case mod_brush:
			R_SetupBrushModel (e);
			break;

		case mod_sprite:
		default:
			break;
		}

		// add it to the draw list
		cl_visedicts[cl_numvisedicts] = e;
		cl_numvisedicts++;
	}
}


qboolean R_CullSphere (float *center, float radius)
{
	if (Vector3Dot (center, frustum[0].normal) - frustum[0].dist <= -radius) return true;
	if (Vector3Dot (center, frustum[1].normal) - frustum[1].dist <= -radius) return true;
	if (Vector3Dot (center, frustum[2].normal) - frustum[2].dist <= -radius) return true;
	if (Vector3Dot (center, frustum[3].normal) - frustum[3].dist <= -radius) return true;

	return false;
}


/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (float *emins, float *emaxs)
{
	int		i;

	for (i = 0; i < 4; i++)
	{
		mplane_t *p = &frustum[i];

		switch (p->signbits)
		{
		default:
		case 0: if (p->normal[0] * emaxs[0] + p->normal[1] * emaxs[1] + p->normal[2] * emaxs[2] < p->dist) return true; break;
		case 1: if (p->normal[0] * emins[0] + p->normal[1] * emaxs[1] + p->normal[2] * emaxs[2] < p->dist) return true; break;
		case 2: if (p->normal[0] * emaxs[0] + p->normal[1] * emins[1] + p->normal[2] * emaxs[2] < p->dist) return true; break;
		case 3: if (p->normal[0] * emins[0] + p->normal[1] * emins[1] + p->normal[2] * emaxs[2] < p->dist) return true; break;
		case 4: if (p->normal[0] * emaxs[0] + p->normal[1] * emaxs[1] + p->normal[2] * emins[2] < p->dist) return true; break;
		case 5: if (p->normal[0] * emins[0] + p->normal[1] * emaxs[1] + p->normal[2] * emins[2] < p->dist) return true; break;
		case 6: if (p->normal[0] * emaxs[0] + p->normal[1] * emins[1] + p->normal[2] * emins[2] < p->dist) return true; break;
		case 7: if (p->normal[0] * emins[0] + p->normal[1] * emins[1] + p->normal[2] * emins[2] < p->dist) return true; break;
		}
	}

	return false;
}


void R_RotateBBox (glmatrix *matrix, const float *inmins, const float *inmaxs, float *outmins, float *outmaxs)
{
	int i, j;

	Vector3Set (outmins, 9999999, 9999999, 9999999);
	Vector3Set (outmaxs, -9999999, -9999999, -9999999);

	// compute a full box
	for (i = 0; i < 8; i++)
	{
		float point[3], transformed[3];

		// get the corner point
		point[0] = (i & 1) ? inmins[0] : inmaxs[0];
		point[1] = (i & 2) ? inmins[1] : inmaxs[1];
		point[2] = (i & 4) ? inmins[2] : inmaxs[2];

		// transform it
		R_Transform (matrix, transformed, point);

		// accumulate to bbox
		for (j = 0; j < 3; j++)
		{
			if (transformed[j] < outmins[j]) outmins[j] = transformed[j];
			if (transformed[j] > outmaxs[j]) outmaxs[j] = transformed[j];
		}
	}
}


qboolean R_CullEntity (entity_t *e)
{
	float emins[3], emaxs[3];

	// compute the correct bbox
	if (e == &cl.viewent)
		return false;
	else if (e->model->type == mod_alias)
	{
		// locate the proper data
		aliashdr_t *hdr = (aliashdr_t *) Mod_Extradata (e->model);
		float mins[3], maxs[3];
		int i;

		// modelgen sets these up so that 0..255 encloses the model volume so here we undo them to get the correct bbox
		// this is light enough to do at runtime :)
		for (i = 0; i < 3; i++)
		{
			mins[i] = hdr->scale_origin[i];
			maxs[i] = mins[i] + hdr->scale[i] * 255;
		}

		R_RotateBBox (&e->localmatrix, mins, maxs, emins, emaxs);
	}
	else if (e->model->type == mod_brush)
		R_RotateBBox (&e->localmatrix, e->model->mins, e->model->maxs, emins, emaxs);
	else return false;

	// now cull the bbox
	return R_CullBox (emins, emaxs);
}


/*
=============================================================

SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *) psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes - 1];

		time = cl.time + currententity->syncbase;

		// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
		// are positive, so we don't have to worry about division by 0
		targettime = time - ((int) (time / fullinterval)) * fullinterval;

		for (i = 0; i < (numframes - 1); i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{	// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1, 1, 1);

	GL_BindPrograms (r_drawtexturedvp, r_drawtexturedfp);
	GL_BindTexture (0, frame->gl_texturenum);

	GL_BlendState (BS_ALPHABLEND);
	GL_DepthState (DS_TESTLESSEQUAL);

	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glEnd ();
}


//==================================================================================

/*
=============
R_DrawEntitiesOnList
=============
*/
void R_DrawEntitiesOnList (void)
{
	int		i;

	if (!r_drawentities.value)
		return;

	// draw sprites seperately, because of alpha blending
	for (i = 0; i < cl_numvisedicts; i++)
	{
		// check for bbculling here
		currententity = cl_visedicts[i];
		currententity->bbculled = R_CullEntity (currententity);

		switch (currententity->model->type)
		{
		case mod_alias:
			R_DrawAliasModel (currententity, r_depthpassmeshvp, r_depthpassfp);
			break;

		case mod_brush:
			R_DrawBrushModel (currententity);
			break;

		default:
			break;
		}
	}
}


void R_DrawSpriteModels (void)
{
	int		i;

	if (!r_drawentities.value)
		return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_sprite:
			R_DrawSpriteModel (currententity);
			break;

		default:
			break;
		}
	}
}


void R_DrawAliasFullbrights (void)
{
	int		i;

	if (!r_drawentities.value)
		return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_alias:
			if (currententity->fb_texturenum)
			{
				GL_BlendState (BS_MAXBLEND);
				GL_DepthState (DS_TESTEQUAL);
				R_DrawAliasModel (currententity, r_fullbrightmeshvp, r_fullbrightfp);
			}

			break;

		default:
			break;
		}
	}
}


/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	if (!r_drawviewmodel.value)
		return;

	if (chase_active.value)
		return;

	if (!r_drawentities.value)
		return;

	if (cl.items & IT_INVISIBILITY)
		return;

	if (cl.stats[STAT_HEALTH] <= 0)
		return;

	currententity = &cl.viewent;

	if (!currententity->model)
		return;

	// hack the depth range to prevent view model from poking into walls
	currententity->effects |= EF_DEPTHHACK;

	R_AddVisedict (currententity);
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value)
		return;
	if (!v_blend[3])
		return;

	GL_BlendState (BS_ALPHABLEND);
	GL_DepthState (DS_NONE);

	glLoadIdentity ();

	glRotatef (-90, 1, 0, 0);	    // put Z going up
	glRotatef (90, 0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	GL_BindPrograms (r_drawcolouredvp, r_drawcolouredfp);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);

	glEnd ();
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j = 0; j < 3; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1 << j;
	}
	return bits;
}


void R_SetFrustum (void)
{
	int i;

	// extract the frustum from the MVP matrix
	// because Quake doesn't cull against the near or far planes we don't need special OpenGL handling for these
	frustum[0].normal[0] = r_mvp_matrix.m4x4[0][3] - r_mvp_matrix.m4x4[0][0];
	frustum[0].normal[1] = r_mvp_matrix.m4x4[1][3] - r_mvp_matrix.m4x4[1][0];
	frustum[0].normal[2] = r_mvp_matrix.m4x4[2][3] - r_mvp_matrix.m4x4[2][0];

	frustum[1].normal[0] = r_mvp_matrix.m4x4[0][3] + r_mvp_matrix.m4x4[0][0];
	frustum[1].normal[1] = r_mvp_matrix.m4x4[1][3] + r_mvp_matrix.m4x4[1][0];
	frustum[1].normal[2] = r_mvp_matrix.m4x4[2][3] + r_mvp_matrix.m4x4[2][0];

	frustum[2].normal[0] = r_mvp_matrix.m4x4[0][3] + r_mvp_matrix.m4x4[0][1];
	frustum[2].normal[1] = r_mvp_matrix.m4x4[1][3] + r_mvp_matrix.m4x4[1][1];
	frustum[2].normal[2] = r_mvp_matrix.m4x4[2][3] + r_mvp_matrix.m4x4[2][1];

	frustum[3].normal[0] = r_mvp_matrix.m4x4[0][3] - r_mvp_matrix.m4x4[0][1];
	frustum[3].normal[1] = r_mvp_matrix.m4x4[1][3] - r_mvp_matrix.m4x4[1][1];
	frustum[3].normal[2] = r_mvp_matrix.m4x4[2][3] - r_mvp_matrix.m4x4[2][1];

	for (i = 0; i < 4; i++)
	{
		Vector3Normalize (frustum[i].normal);
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = Vector3Dot (r_refdef.vieworigin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
	// don't allow cheats in multiplayer
	if (cl.maxclients > 1)
		Cvar_Set ("r_fullbright", "0");

	R_AnimateLight (cl.time);

	r_framecount++;

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworigin, r_origin);

	// AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;
}


/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	glViewport (0, sb_lines, vid.width, vid.height - sb_lines);

	R_IdentityMatrix (&r_projection_matrix);
	R_FrustumMatrix (&r_projection_matrix, r_refdef.fov_x, r_refdef.fov_y);

	R_IdentityMatrix (&r_world_matrix);
	R_CameraMatrix (&r_world_matrix, r_refdef.vieworigin, r_refdef.viewangles);

	glMatrixMode (GL_PROJECTION);
	glLoadMatrixf (r_projection_matrix.m16);

	glMatrixMode (GL_MODELVIEW);
	glLoadMatrixf (r_world_matrix.m16);

	// derive a full MVP (the initial "M" is going to be identity so we can ignore it for now)
	R_MultMatrix (&r_mvp_matrix, &r_world_matrix, &r_projection_matrix);

	// take these from the view matrix
	Vector3Set (vpn,   -r_world_matrix.m4x4[0][2], -r_world_matrix.m4x4[1][2], -r_world_matrix.m4x4[2][2]);
	Vector3Set (vup,    r_world_matrix.m4x4[0][1],  r_world_matrix.m4x4[1][1],  r_world_matrix.m4x4[2][1]);
	Vector3Set (vright, r_world_matrix.m4x4[0][0],  r_world_matrix.m4x4[1][0],  r_world_matrix.m4x4[2][0]);

	// set drawing parms
	glCullFace (GL_FRONT);

	if (gl_cull.value)
		glEnable (GL_CULL_FACE);
	else glDisable (GL_CULL_FACE);

	R_SetupSkyShaderParams ();
	R_SetupWarpShaderParams ();
}


void R_RenderDepthPrePass (void)
{
	gldepthmin = 0;
	gldepthmax = 1;

	glDepthRange (gldepthmin, gldepthmax);

	// in OpenGL depth-writing state affects clears so we must enable depth writing before we can clear
	// somebody needs a really good slap about the head for this
	GL_BlendState (BS_NONE);
	GL_DepthState (DS_WRITE | DS_TESTLESSEQUAL);

	// NOW it's OK to clear
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	// colormask also affects clears so we cannot disable color writes until AFTER we have cleared
	glColorMask (GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	R_DrawWorld ();		// adds static entities to the list

	R_DrawViewModel ();	// just adds view model to cl_visedicts

	R_DrawEntitiesOnList ();

	glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
}


/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	if (gl_finish.value)
		glFinish ();

	// render normal view
	R_SetupFrame ();

	R_SetupGL ();

	R_SetFrustum ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_RenderDepthPrePass ();

	R_RenderLightsAndShadows ();

	R_DrawExtraSurfaces ();

	R_DrawAliasFullbrights ();

	R_DrawSpriteModels ();

	R_DrawParticles ();

	R_PolyBlend ();
}

